package laboratoire4.strategies;

import laboratoire4.Action;
import laboratoire4.game.Game;
import laboratoire4.game.GameUtils;
import laboratoire4.game.SimulatedGame;
import laboratoire4.pawns.PawnMovement;
import laboratoire4.pawns.SimulatedPawn;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

public abstract class MiniMaxStrategy implements Strategy {
    private static final int MAX_DEPTH = 4;

    protected SimulatedGame game;

    @Override
    public EvaluationResult getNextMove(Game game) {
        return miniMax(game);
    }

    private EvaluationResult miniMax(Game game) {
        this.game = new SimulatedGame(game.getBoard(), game.getPlayer());

        EvaluationResult maxResult = null;
        int maxScore = Integer.MIN_VALUE;

        for (Action<SimulatedPawn> action : getActions(true)) {
            int score = min(0, Integer.MIN_VALUE, Integer.MAX_VALUE);

            if (maxResult == null || score > maxScore) {
                SimulatedPawn pawn = action.getPawn();
                maxResult = new EvaluationResult(pawn.getRow(), pawn.getCol(), action.getMovement());
            }
        }

        return maxResult;
    }

    private int max(int depth, int alpha, int beta) {
        if (depth >= MAX_DEPTH) {
            return evaluateSimulation();
        }

        int maxScore = Integer.MIN_VALUE;

        for (Action<SimulatedPawn> action : getActions(true)) {
            SimulatedPawn pawn = action.getPawn();
            PawnMovement movement = action.getMovement();

            game.move(pawn, movement);
            int score = min(depth + 1, Math.max(alpha, maxScore), beta);
            game.revertMove();

            if (score > maxScore) {
                maxScore = score;
            }

            if (maxScore >= beta) {
                return maxScore;
            }
        }

        return maxScore;
    }

    private int min(int depth, int alpha, int beta) {
        if (depth >= MAX_DEPTH) {
            return evaluateSimulation();
        }

        int minScore = Integer.MAX_VALUE;

        for (Action<SimulatedPawn> action : getActions(false)) {
            SimulatedPawn pawn = action.getPawn();
            PawnMovement movement = action.getMovement();

            game.move(pawn, movement);
            int score = max(depth + 1, alpha, Math.min(beta, minScore)) * -1;
            game.revertMove();

            if (score < minScore) {
                minScore = score;
            }

            if (minScore <= alpha) {
                return minScore;
            }
        }

        return minScore;
    }

    private Collection<Action<SimulatedPawn>> getActions(boolean max) {
        List<Action<SimulatedPawn>> actions = new ArrayList<>();
        Collection<SimulatedPawn> pawns = max ? getMaxPawns() : getMinPawns();
        PawnMovement[] movements = PawnMovement.values();

        for (SimulatedPawn pawn : pawns) {
            for (PawnMovement movement : movements) {
                if (pawn.isMoveValid(game.getBoard(), movement)) {
                    actions.add(new Action<>(pawn, movement));
                }
            }
        }

        Collections.shuffle(actions);
        return actions;
    }

    protected Collection<SimulatedPawn> getMaxPawns() {
        return GameUtils.getMaxPawns(game.getBoard(), game.getPlayer());
    }

    protected Collection<SimulatedPawn> getMinPawns() {
        return GameUtils.getMinPawns(game.getBoard(), game.getPlayer());
    }

    protected abstract int evaluateSimulation();
}
